\ jim

\ ' (TICK) behaves non-standardly -- it parses the next word at
\ compile time.  Instead it should parse at runtime, and ['] should
\ parse at compile time.  (['] is defined in jonesforth.fs,
\ but doesn't parse the next word -- instead it transforms itself
\ into "LIT" and relies on the compiler to compile the next word.
\ Problem is that next word may update LATESTCOL when compiled,
\ which will break the TCO check.)

: '  WORD FIND   DUP 0= IF ." undefined word" ABORT THEN   >CFA ;
: POSTPONE ' , ; IMMEDIATE         \ not working correctly
: [']  '  POSTPONE LITERAL ; immediate
: [CHAR] CHAR POSTPONE LITERAL ; IMMEDIATE


\ ------------------- tail recursion
\ notes: HERE broken--jonesforth needs HERE @ for here pointer, should be HERE
\ added REVEAL to asm and make ; use it instead of toggling HIDDEN flag
\ PREVCALL? is not robust, as prevxt may be data (ex. LITERAL, .") -- esp. because DOCOL == 0
\   example, in : foo 1 2 3 ['] . -; ;
\   -> which should be : foo 1 2 3 22316 ;
\   the words translate to LIT 1 LIT 2 LIT 3 LIT 'BRANCH <OFFSET>
\   because the XT of . in the literal (22316) is looked up and transformed into a BRANCH
: RECURSIVE  LATEST @ REVEAL ; IMMEDIATE
: DP HERE ;
: <RESOLVE DP @ - , ;      \ BRANCH is relative
: PREVXT DP @ 4- @ ;

: PREVCALL? PREVXT @ DOCOL = ;
: TCO! ( xt -- ) -4 DP +!   ['] BRANCH ,   >PFA <RESOLVE ;     \ bypass DOCOL's stack push
: TRY-TCO PREVCALL? IF PREVXT TCO! TRUE ELSE FALSE THEN ;
    
: -; TRY-TCO  0= IF ." -; expected call " ABORT THEN ; IMMEDIATE
: x; TRY-TCO  0= IF ['] EXIT , THEN ; IMMEDIATE
: ;; TRY-TCO  DROP  [COMPILE] ;  ; IMMEDIATE

\ correctly conserves stack space
: TREC RECURSIVE
    DUP 0= IF ." DONE " DROP x; THEN RSP@ . 1- TREC -; ;
: TREC2 RECURSIVE   \ count down then tail-call into TREC
    DUP 0= IF ." DONE " DROP  5 TREC x; THEN RSP@ . 1- TREC2 ;;
\ non-tail-recursive--crashes at around depth 1000
: REC RECURSIVE
    DUP 0= IF ." DONE " DROP x; THEN RSP@ . 1- REC ;


\ ---- using LATESTCOL, a slight modification by me to the INTERPRET loop
\ this var contains latest non-primitive word (a colon definition) compiled in INTERPRET 
\ more robust
\ note: load above first

: PREVCALL? DP @ 4- LATESTCOL @ = ;           \ Only difference with above
: TRY-TCO PREVCALL? IF ." tco ok" PREVXT TCO! TRUE ELSE FALSE THEN ;
: -; TRY-TCO  0= IF ." -; expected call " ABORT THEN ; IMMEDIATE
: x; TRY-TCO  0= IF ['] EXIT , THEN ; IMMEDIATE
: ;; TRY-TCO  DROP  [COMPILE] ;  ; IMMEDIATE

\ latestcol
\ : FOO 1 2 3 ['] . -; ;    \ will abort now
: BAR 1 2 3 ['] . ;;        \ will correctly compile literal

\ ------------------------------------------------

: CELL+ 1 CELLS + ;
: 2@  DUP SWAP CELL+ @ SWAP @ ;
